import scipy.spatial.distance as spd


class DistanceType:

    def eucos(self):
        return 'eucos'

    def euclidean(self):
        return 'euclidean'

    def cosine(self):
        return 'cosine'

    def squeuclidean(self):
        return 'squeuclidean'


def compute_distance(query_channel, channel, mean_vec, distance_type = 'eucos'):
    """ Compute the specified distance type between chanels of mean vector and query image.
    In caffe library, FC8 layer consists of 10 channels. Here, we compute distance
    of distance of each channel (from query image) with respective channel of
    Mean Activation Vector. In the paper, we considered a hybrid distance eucos which
    combines euclidean and cosine distance for bouding open space. Alternatively,
    other distances such as euclidean or cosine can also be used.

    Input:
    --------
    query_channel: Particular FC8 channel of query image
    channel: channel number under consideration
    mean_vec: mean activation vector
    Output:
    --------
    query_distance : Distance between respective channels
    """

    if distance_type == 'eucos':
        query_distance = spd.euclidean(mean_vec[channel, :], query_channel) /200. + spd.cosine(mean_vec[channel, :], query_channel)
    elif distance_type == 'euclidean':
        query_distance = spd.euclidean(mean_vec[channel, :], query_channel) /200.
    elif distance_type == 'cosine':
        query_distance = spd.cosine(mean_vec[channel, :], query_channel)
    elif distance_type == DistanceType.squeuclidean():
        query_distance = spd.sqeuclidean(mean_vec[channel, :], query_channel)
    else:
        print("distance type not known: enter either of eucos, euclidean or cosine")
    return query_distance


import numpy as np
import libmr

def weibull_tailfitting(mean, distance, labellist, tailsize=20, distance_type='eucos'):
    weibull_model = {}
    for category in labellist:
        weibull_model[category] = {}
        distance_scores = np.array(distance[category][distance_type])
        meantrain_vec = np.array(mean[category])
        weibull_model[category]['distances_%s' % distance_type] = distance_scores
        weibull_model[category]['mean_vec'] = meantrain_vec
        weibull_model[category]['weibull_model'] = []
        mr = libmr.MR()
        tailtofit = sorted(distance_scores)[-tailsize:]
        mr.fit_high(tailtofit, len(tailtofit))
        weibull_model[category]['weibull_model'] += [mr]
    return weibull_model


def query_weibull(category_name, weibull_model, distance_type='eucos'):
    category_weibull = []
    category_weibull += [weibull_model[category_name]['mean_vec']]
    category_weibull += [weibull_model[category_name]['distances_%s' % distance_type]]
    category_weibull += [weibull_model[category_name]['weibull_model']]

    return category_weibull



from EVT_fitting import*
import scipy as sp
from openmax_utils import compute_distance
import sys
import numpy as np


def computeOpenMaxProbability(openmax_fc8, openmax_score_u, classes=10, channels=1):

    """ Convert the scores in probability value using openmax

    Input:
    ---------------
    openmax_fc8 : modified FC8 layer from Weibull based computation
    openmax_score_u : degree
    Output:
    ---------------
    modified_scores : probability values modified using OpenMax framework,
    by incorporating degree of uncertainity/openness for a given class

    """
    prob_scores, prob_unknowns = [], []
    for channel in range(channels):
        channel_scores, channel_unknowns = [], []
        for category in range(classes):
            channel_scores += [sp.exp(openmax_fc8[channel, category])]

        total_denominator = sp.sum(sp.exp(openmax_fc8[channel, :])) + sp.exp(sp.sum(openmax_score_u[channel, :]))
        if total_denominator is np.inf:
            total_denominator = sys.float_info.max

        prob_scores += [channel_scores / total_denominator]
        prob_unknowns += [sp.exp(sp.sum(openmax_score_u[channel, :])) / total_denominator]

    prob_scores = sp.asarray(prob_scores)
    prob_unknowns = sp.asarray(prob_unknowns)

    scores = sp.mean(prob_scores, axis=0)
    unknowns = sp.mean(prob_unknowns, axis=0)
    modified_scores = scores.tolist() + [unknowns]
    assert len(modified_scores) == (classes + 1)
    return modified_scores


def recalibrate_scores(weibull_model, labellist, imgarr,
                       layer='fc8', alpharank=10, distance_type='eucos', classes=10, channels=1):
    """
    Given FC8 features for an image, list of weibull model for each class,
    re-calibrate scores
    Input:
    ---------------
    weibull_model : pre-computed weibull_model obtained from weibull_tailfitting() function
    labellist : ImageNet 2012 labellist
    imgarr : features for a particular image extracted using caffe architecture

    Output:
    ---------------
    openmax_probab: Probability values for a given class computed using OpenMax
    softmax_probab: Probability values for a given class computed using SoftMax (these
    were precomputed from caffe architecture. Function returns them for the sake
    of convienence)
    """
    if alpharank > len(labellist):
        alpharank = len(labellist)
    imglayer = imgarr[layer]
    ranked_list = np.argsort(imgarr['scores']).ravel()[::-1]
    alpha_weights = [((alpharank + 1) - i) / float(alpharank) for i in range(1, alpharank + 1)]
    ranked_alpha = sp.zeros(1000)
    for i in range(len(alpha_weights)):
        ranked_alpha[ranked_list[i]] = alpha_weights[i]

    # Now recalibrate each fc8 score for each channel and for each class
    # to include probability of unknown
    openmax_fc8, openmax_score_u = [], []
    for channel in range(channels):
        channel_scores = imglayer[channel, :]
        openmax_fc8_channel = []
        openmax_fc8_unknown = []
        count = 0
        for categoryid in range(classes):
            # get distance between current channel and mean vector
            category_weibull = query_weibull(labellist[categoryid], weibull_model, distance_type=distance_type)
            channel_distance = compute_distance(channel_scores, channel, category_weibull[0],
                                                distance_type=distance_type)

            # obtain w_score for the distance and compute probability of the distance
            # being unknown wrt to mean training vector and channel distances for
            # category and channel under consideration
            wscore = category_weibull[2][channel].w_score(channel_distance)
            modified_fc8_score = np.ravel(channel_scores)[categoryid] * (1 - wscore * ranked_alpha[categoryid])
            openmax_fc8_channel += [modified_fc8_score]
            openmax_fc8_unknown += [np.ravel(channel_scores)[categoryid] - modified_fc8_score]

        # gather modified scores fc8 scores for each channel for the given image
        openmax_fc8 += [openmax_fc8_channel]
        openmax_score_u += [openmax_fc8_unknown]
    openmax_fc8 = sp.asarray(openmax_fc8)
    openmax_score_u = sp.asarray(openmax_score_u)

    # Pass the recalibrated fc8 scores for the image into openmax
    openmax_probab = computeOpenMaxProbability(openmax_fc8, openmax_score_u, classes=classes)
    softmax_probab = imgarr['scores'].ravel()
    return sp.asarray(openmax_probab), sp.asarray(softmax_probab)


from typing import Dict, Any, Union

import numpy as np
from keras import backend as K
import scipy.spatial.distance as spd
import os

from numpy.core._multiarray_umath import ndarray
from tensorflow.keras.models import load_model
from EVT_fitting import weibull_tailfitting
from compute_openmax import recalibrate_scores
from tqdm import tqdm


class Openmax:
    """Class for openmax computations.
    Constructor needs a pretrained model with 2 named layers. One should represent the activation vectors and the other
    contains the softmax scores.  The input data and labels as a tuple, the input shape of the first layer,
    optional arguments that represent the observed named layers of the model(eg. dense_01, FNC7, softmax...)
    The layer names have to match the layer names of the model
    """
    DATA_PATH = "data/openmax/"
    MODEL_PATH = "model/openmax/"

    def __init__(self, model, x_data, y_data, input_shape: tuple, name: str, overwrite=False, *args):
        self.threshold = 0.5
        self.weibull_model = None
        self.input_shape = (1,) + input_shape
        self.model = model
        self.x_data = x_data
        self.y_data = y_data
        self.name = name
        self.observations = []
        self.backend_functions = []
        self.CLASSES = np.unique(self.y_data)
        self.overwrite = overwrite
        for observation in args:
            self.observations.append(observation)
            self.init_backend_functions(observation)
        if not os.path.exists(self.DATA_PATH+self.name+"/"):
            os.makedirs(self.DATA_PATH+self.name+"/")
        if not os.path.exists(self.MODEL_PATH):
            os.makedirs(self.MODEL_PATH)

    def save_model(self):
        self.model.save(self.MODEL_PATH + self.name+".h5")

    def load_model(self, custom_objects=None):
        self.model = load_model(self.MODEL_PATH + self.name+".h5", custom_objects=custom_objects)
        return self.model

    def get_correctly_classified(self):
        predictions = self.model.predict(self.x_data)
        out = []
        for idx, p in enumerate(predictions):
            p_label = np.argmax(p)
            if p_label == self.y_data[idx] and np.amax(p) > self.threshold:
                out.append(True)
            else:
                out.append(False)
        percentage = sum(out)/len(self.x_data)
        print("{} percent of the data were predicted correctly and are considered for the probability distribution.".format(percentage*100))
        return np.array(out)

    def init_backend_functions(self, layer):
        if type(layer) is str:
            self.backend_functions.append(K.function([self.model.layers[0].input, K.learning_phase()],
                                         [self.model.get_layer(layer).output]))
        else:
            self.backend_functions.append(K.function([self.model.layers[0].input, K.learning_phase()],
                                         [self.model.layers[layer].output]))

    def get_backend_function(self, layer, input):
        return self.backend_functions[self.observations.index(layer)]([input, 0])[0]

    def compute_feature(self, image_set):
        if image_set.shape == self.input_shape:
            return np.asarray(self.get_backend_function(self.observations[0], image_set))
        out = []
        for image in image_set:
            out.append(self.get_backend_function(self.observations[0], image.reshape(self.input_shape)))
        return np.asarray(out)

    def compute_score(self, image_set):

        if image_set.shape == self.input_shape:
            return np.asarray(self.get_backend_function(self.observations[1], image_set))
        out = []
        for image in image_set:
            out.append(self.get_backend_function(self.observations[1], image.reshape(self.input_shape)))
        return np.asarray(out)

    def compute_means(self, feature):
        return np.mean(feature, axis=0)

    def compute_distance(self, mean_train_vector, feature, category_name):
        eu_dist, cos_dist, eucos_dist = [], [], []
        for feat in feature:
            eu_dist += [spd.euclidean(mean_train_vector, feat)]
            cos_dist += [spd.cosine(mean_train_vector, feat)]
            eucos_dist += [spd.euclidean(mean_train_vector, feat) / 200. +
                           spd.cosine(mean_train_vector, feat)]
        distances = {'eucos': eucos_dist, 'cosine': cos_dist, 'euclidean': eu_dist}
        return distances

    def prepare_data(self, x, y):
        index = np.argsort(y)
        return x[index], y[index]

    def sort_data(self):
        index = self.get_correctly_classified()
        x_data = self.x_data[index]
        y_data = self.y_data[index]
        sorted_x, sorted_y = self.prepare_data(x_data, y_data)
        return sorted_x, sorted_y

    def compute_data(self):
        print("Started Data Computation")
        x_data, y_data = self.sort_data()
        all_means = []
        all_distances = []
        for i in tqdm(range(len(self.CLASSES))):
            feat = self.compute_feature(x_data[y_data == i])
            mean = self.compute_means(feat)
            distance = self.compute_distance(mean, feat, i)
            all_means.append(mean)
            all_distances.append(distance)
            np.save(self.DATA_PATH + self.name + "/" + 'activation_vector_class' + str(i), feat)
        np.save(self.DATA_PATH + self.name + "/" + 'mean', all_means)
        np.save(self.DATA_PATH + self.name + "/" + 'distance', all_distances)
        print("Finished Data Computation")

    def set_threshold(self, threshold: float):
        self.threshold = threshold

    def create_weibull(self, distance_type: str, tailsize=20, overwrite=False):
        if overwrite:
            self.compute_data()
        elif len(os.listdir(self.DATA_PATH + self.name+"/")) == 0:
            self.compute_data()
        means = np.load(self.DATA_PATH + self.name + "/" + 'mean.npy', allow_pickle=True)
        distances = np.load(self.DATA_PATH + self.name + "/" + 'distance.npy', allow_pickle=True)
        self.weibull_model = weibull_tailfitting(means, distances, range(len(self.CLASSES)),
                                                 distance_type=distance_type, tailsize=tailsize)
        return self.weibull_model

    def predict(self, image, distance_type='eucos', alpha=10):
        if self.weibull_model is None:
            self.create_weibull(distance_type, self.overwrite)
        if not image.shape == self.input_shape:
            image = image.reshape(self.input_shape)
        imgarr = {'scores': self.compute_score(image),
                      self.observations[0]: self.compute_feature(image)}
        openmax, softmax = recalibrate_scores(self.weibull_model, range(len(self.CLASSES)), imgarr,
                                              self.observations[0], distance_type=distance_type,
                                              classes=len(self.CLASSES),alpharank=alpha)
        return openmax, softmax

    def adapt_tailsize(self, distance_type='eucos', tail_range=range(0,40), alpha_range=range(0,10)):
        adaption = np.zeros((len(alpha_range) + 1, len(tail_range) + 1))
        for image in tqdm(self.x_data):
            for alpha in alpha_range:
                for tail in tail_range:
                    self.create_weibull(distance_type, tailsize=tail, overwrite=False)
                    if not image.shape == self.input_shape:
                        image = image.reshape(self.input_shape)
                    imgarr = {'scores': self.compute_score(image),
                              self.observations[0]: self.compute_feature(image)}
                    openmax, softmax = recalibrate_scores(self.weibull_model, range(len(self.CLASSES)), imgarr,
                                                          self.observations[0], distance_type=distance_type,
                                                          classes=len(self.CLASSES), alpharank=alpha)
                    if np.argmax(openmax) == np.argmax(softmax):
                        adaption[alpha][tail] += 1
        return adaption



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="keras-openmax",
    version="1.0.2",
    author="Clemens Brackmann",
    author_email="clemens.brackmann@web.de",
    description="Keras implementation of Openmax",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DietmarKracht/Keras-Openmax",
    py_modules=["Openmax", "compute_openmax", "EVT_fitting", "openmax_utils"],
    install_requires=[
        'Cython>=0.29.15',
        'Keras==2.3.1',
        'libmr>=0.1.9',
        'matplotlib>=3.2.0',
        'numpy>=1.18.1',
        'Pillow>=7.0.0',
        'scipy==1.4.1',
        'tensorflow==2.1.0',
        'tqdm>=4.43.0',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)